/*
 * use xtest extn to get keybd events and check for overuse.
 * Mark M Martin.  cetia. @(#)restwatch.c	1.13 95/06/06
 */
/*
 * define USEXTESTEXTN1 if you can compile with it.  Though having
 * this in the libXext.a does not mean it is supported in the server.
 * It's the best cos the most efficient and also gets you mouse clicks.
 */
#define USEXTESTEXTN1

char *ProgramName;
#include <stdio.h>
#include <X11/Xos.h>
#include <X11/Xmd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Form.h>

/*
 * usea2xhack is true if a2x being used for kbd and has been hacked to
 * generate a sendevent on the root window, since otherwise this program
 * gets no input from XTestExtension1 for these events.
 * If you dont have this hack it should matter if it is set true.
 */
# ifdef USEXTESTEXTN1
#include <X11/extensions/xtestext1.h>
extern int XTestInputActionType;
int usea2xhack = TRUE;
int selectallinput = FALSE;	/* no point if XTestExtension1 */
# else
int XTestInputActionType = -99;	/* just a dummy */
int usea2xhack = FALSE;	/* no point if no XTestExtension1 */
int selectallinput = TRUE;
# endif

#define IDLING	(1)
#define WORKING	(2)
#define RESTING	(3)
char *statename[] = { 0, "idle", "busy", "resting" };

char *usagestr = "\
usage: %s -display display\n\
	-busy	busytime (secs)\n\
	-idle	idletime\n\
	-verbose show events and times\n\
	-keystrokes count	update display count\n\
	-continuous msecs	max delay between keystrokes\n\
	-log minutes		log totals and zero after this idletime\n\
	-nomap	dont force up window\n\
	-unmap	force iconification of window\n\
	-noext	dont use xtestextension1\n\
";

static void usage ()
{
    fprintf (stderr, usagestr, ProgramName);
    exit (1);
}

typedef struct {
    int		maxbusy;
    int		minidle;
    int		keystrokes;
    int		updatetime;
    int		continuous;
    int		log;
    Boolean	verbose;
    Boolean	tomap;
    Boolean	tounmap;
    Boolean	useext;
    Pixel	busyfg;
    Pixel	idlefg;
    Pixel	recoverfg;
} myResources;

/*
 * Command line options table.  Only resources are entered here...there is a
 * pass over the remaining options after XtParseCommand is let loose. 
 */

static XrmOptionDescRec  cmdoptions[] = {
{"-busy",	".busy",	XrmoptionSepArg},
{"-idle",	".idle",	XrmoptionSepArg},
{"-keystrokes",	".keystrokes",	XrmoptionSepArg},
{"-continuous",	".continuous",	XrmoptionSepArg},
{"-log",	".log",		XrmoptionSepArg},
{"-verbose",	".verbose",	XrmoptionNoArg, "true"},
{"-nomap",	".map",		XrmoptionNoArg, "false"},
{"-unmap",	".unmap",	XrmoptionNoArg, "true"},
{"-noext",	".ext",		XrmoptionNoArg, "false"},
};

/*
 * The structure containing the resource information for the application resources.
 */

#define Offset(field) (XtOffset(myResources *, field))

static XtResource my_resources[] = {
  {"verbose", XtCBoolean, XtRBoolean, sizeof(Boolean),
     Offset(verbose), XtRImmediate, (caddr_t) FALSE},
  {"map", XtCBoolean, XtRBoolean, sizeof(Boolean),
     Offset(tomap), XtRImmediate, (caddr_t) TRUE},
  {"unmap", XtCBoolean, XtRBoolean, sizeof(Boolean),
     Offset(tounmap), XtRImmediate, (caddr_t) FALSE},
  {"ext", XtCBoolean, XtRBoolean, sizeof(Boolean),
     Offset(useext), XtRImmediate, (caddr_t) TRUE},
  {"busy", XtCValue, XtRInt, sizeof(int),
     Offset(maxbusy), XtRImmediate, (caddr_t) 180},
  {"time", XtCValue, XtRInt, sizeof(int),
     Offset(updatetime), XtRImmediate, (caddr_t) 60},
  {"keystrokes", XtCValue, XtRInt, sizeof(int),
     Offset(keystrokes), XtRImmediate, (caddr_t) 100},
  {"continuous", XtCValue, XtRInt, sizeof(int),
     Offset(continuous), XtRImmediate, (caddr_t) 2000},
  {"log", XtCValue, XtRInt, sizeof(int),
     Offset(log), XtRImmediate, (caddr_t) 360},
  {"idle", XtCValue, XtRInt, sizeof(int),
     Offset(minidle), XtRImmediate, (caddr_t) 30},
  {"busyFg", XtCForeground, XtRPixel, sizeof(Pixel),
     Offset(busyfg), XtRString, (caddr_t) "yellow"},
  {"idleFg", XtCForeground, XtRPixel, sizeof(Pixel),
     Offset(idlefg), XtRString, (caddr_t) "black"},
  {"recoverFg", XtCForeground, XtRPixel, sizeof(Pixel),
     Offset(recoverfg), XtRString, (caddr_t) "red"},
};

/*XtActionsRec myactions[] = {};*/
XtActionsRec myactions[];

String fallback[] = {
	"*resizable:false",
	"*label:",
	"*top:ChainTop",
	"*bottom:ChainTop",
	"*label.bottom:ChainBottom",
	"*left:ChainLeft",
	"*right:ChainRight",
	"*horizDistance:0",
	"*vertDistance:0",
	"*borderWidth:0",
	"*tottime.fromVert:time",
	"*keystrokes.fromVert:tottime",
	"*label.fromVert:keystrokes",
	".?.geometry:200x100-190+1",
	NULL
};

XtAppContext app;
myResources myres;
Widget toplevel,form,label,timew,tottimew,keystrokesw;
Display *display;

int usingtod = FALSE;
int currentstate = IDLING;
struct timeval tod = {0,0};
char *actstr;
int repolltime = 0;
#define REPOLLTIME	(60*1000)	/* msecs. very rough */
#define WANTEDEVENTS KeyPressMask /* only 1 can select on ButtonPressMask */

/*
 * Unfortunately, scanning the window tree is not an atomic affair.
 * If a window disappears half way throught you can get bad window errors
 * that will kill the application. trap these and ignore them
 */
static int (*realErrorHandler)();
static int windowErrorsPossible = 0;
static int checkError(display, error)
Display *display;
XErrorEvent *error;
{
    if(windowErrorsPossible && error->error_code==BadWindow)return;
    (realErrorHandler)(display, error);
    /* should never return here */
    exit(1);
}

/*
 * internally, times are help in milliseconds
 */
int
main(argc, argv)int argc;char *argv[];
{
    Arg args[10];
    Cardinal n;
    int i,fds,rdfds;
    int event_base, error_base;
    int major_version, minor_version;
    struct timeval timeout;

    ProgramName = argv[0];
    setbuf(stdout,NULL);
    n = 0;
    toplevel = XtAppInitialize(&app, "Restwatch",
			    cmdoptions, XtNumber(cmdoptions), &argc, argv,
			       fallback, args, n);
      
    XtGetApplicationResources( toplevel, (caddr_t) &myres, 
			      my_resources, XtNumber(my_resources),
			      NULL, (Cardinal) 0);
    argc--, argv++;
    if(argc!=0)
	    usage();
    
#if 0
    XtAppAddActions(app,myactions,sizeof(myactions)/sizeof(myactions[0]));
#endif
    display = XtDisplay(toplevel);
    n = 0;
    form = XtCreateManagedWidget ("form", formWidgetClass,
				 toplevel, args, n);
    timew = XtCreateManagedWidget ("time", labelWidgetClass,
				 form, args, n);
    tottimew = XtCreateManagedWidget ("tottime", labelWidgetClass,
				 form, args, n);
    keystrokesw = XtCreateManagedWidget ("keystrokes", labelWidgetClass,
				 form, args, n);
    label = XtCreateManagedWidget ("label", labelWidgetClass,
				 form, args, n);
    XtRealizeWidget (toplevel);

    myres.minidle *= 1000;	/* secs to millisecs */
    myres.maxbusy *= 1000;
    myres.log *= 60*1000;	/* minutes to millisecs */
    timeout.tv_sec = myres.minidle/1000;
    timeout.tv_usec = 0;
    fds = 1<<ConnectionNumber(display);
#ifdef USEXTESTEXTN1
    if(!myres.useext || XTestGetInput(display)!=0) {
	if(myres.useext)
	    fprintf(stderr,"%s: %s extension not supported on server \"%s\"\n",
		    ProgramName, XTestEXTENSION_NAME,DisplayString(display));
	XTestInputActionType = -99;	/* just a dummy */
	usea2xhack = FALSE;
	selectallinput = TRUE;
    }
#endif
    /* if usea2xhack then get sendevents on root window */
    if(usea2xhack || selectallinput)
	XSelectInput(display, DefaultRootWindow(display), KeyPressMask);
    if(selectallinput)realErrorHandler = XSetErrorHandler(checkError);

    /* replaces XtAppMainLoop(app); */
    while(1){
	XEvent event;
	int rc;
		
	if(selectallinput)
	    if(repolltime<=0){
		XSync(display,FALSE); /* handle genuine errors */
		windowErrorsPossible = TRUE;
		selectAllWindows(DefaultRootWindow(display));
		XSync(display,FALSE); /* handle spurious errors */
		windowErrorsPossible = FALSE;
		repolltime = REPOLLTIME;
	    }
	/* despite there being several time periods, we only ever
	 * change state at a fixed time, i.e. without being provoked
	 * by an input event, when going from resting (forbid) to idle (ok).
	 */
	rdfds = fds;
	if(currentstate!=IDLING || selectallinput)
	    /* cannot hang in read: need to go back to idle, or poll windows */
	    if(!XtAppPending(app)){	/* no input ready */
		rc = (int) select(32,(fd_set *)&rdfds,0,0,&timeout);
		if(rc<0)pexit("select");
		if(rc==0){
		    /* timeout on select. do reset with dummy event */
		    actstr = "delay";
		    if(usingtod)gettimeofday(&tod,0);
		    notetime(timeout.tv_sec*1000,TRUE);
		    continue;
		}
	    }
	XtAppNextEvent(app, &event);
	if (!handleevent(&event)){
	    /* normal toolkit events, like window refresh */
	    XtDispatchEvent(&event);
	    continue;
	}
    }
}
/* sendevents from a2x have no delay so must use tod.
 * but cannot mix with delays from xtest cos they dont know
 * about sendevent, so all use tod afer 1st sendevent.
 * return true if event was understood.
 */
handleevent(eventp)
XEvent *eventp;
{
    int timepassed,ignore;
    struct timeval newtod;
    static int lasteventtime = 0;	

    if(usingtod){
	gettimeofday(&newtod,0);
	timepassed = (newtod.tv_sec-tod.tv_sec)*1000+
	    (newtod.tv_usec-tod.tv_usec)/1000;
	tod = newtod;
    }
    ignore = FALSE;
#ifdef USEXTESTEXTN1
    if(eventp->type==XTestInputActionType){
	if(usingtod)interpxtestaction(eventp,&ignore);
	else timepassed = interpxtestaction(eventp,&ignore);
    }else
#endif
    if(eventp->xany.send_event){
	if(usea2xhack){
	    /* sendevent timestamp junk so must start using gettimeofday */
	    if(!usingtod){
		usingtod = TRUE;
		gettimeofday(&tod,0);
		timepassed = 0;
	    }
	}else{ /* else ignore sendevent */
	    return FALSE;
	}
	actstr = "se";
    }else if(eventp->type==KeyPress){
	actstr = "key";
	if(!usingtod){
	    if(lasteventtime==0)timepassed = 0;/* first event will be huge */
	    else timepassed = eventp->xkey.time-lasteventtime;
	    lasteventtime = eventp->xkey.time;
	}
    }else{
	if(usingtod){
	    actstr = "other";
	    notetime(timepassed,TRUE);
	}
	return FALSE;	/* not for me */
    }
    notetime(timepassed,ignore);
    return TRUE;	/* event understood */
}
/*
 * accumlate time in each state until pass limit and change state
 */
notetime(timepassed,ignore)
int timepassed,ignore;
{
    static int totstatetime, laststatetime, tottime, lasttottime, totcount;
    int newstate,t;

    if(myres.verbose)
	printf("%s %d %s %dms %d\n",actstr,currentstate,
	       ignore?"ignored":"",timepassed,totstatetime/1000);
    repolltime -= timepassed;
    totstatetime += timepassed;
    newstate = currentstate;
    if(!ignore){
	if(timepassed<=myres.continuous)tottime += timepassed;
	t = (tottime/1000)/myres.updatetime;
	if(t!=lasttottime){
	    lasttottime = t;
	    updatetypingtime(tottime/1000);
	}
	totcount++;
	if(selectallinput){
	    if((totcount%myres.keystrokes)==0)
		updatekeystrokes(totcount);
	}else if(((totcount/2)%myres.keystrokes)==0)
	    updatekeystrokes(totcount/2);	/* counts downs and ups */
    }
    switch(currentstate){
    case IDLING:
	/* add in resting time */
	totstatetime += laststatetime;
	laststatetime = 0;
	if(!ignore)newstate = WORKING;
    /* reset after 6 hrs idle */
	if(totstatetime>myres.log){
	    time_t tod = time((time_t *)0);
	    char *atod = ctime(&tod);
	    atod[sizeof("Sun Sep 16 01:03")-1] = '\0';
	    if(!selectallinput)totcount /= 2;
	    tottime /= 1000;	/* millisecs to secs */
	    if(totcount!=0 && tottime!=0)
		printf("%s total %d keystrokes  %d:%02d hrs:min  %d/min\n",
		       atod, totcount,tottime/3600,(tottime/60)%60,
		       totcount*60/tottime);
	    totcount = tottime = 0;
	}
	break;
    case WORKING:
	if(ignore){
	    if(timepassed>=myres.minidle){
		/* skip resting, time already passed */
		totstatetime = timepassed-myres.minidle;
		currentstate = RESTING;
		newstate = IDLING;
	    }
	}else
	    if(totstatetime>=myres.maxbusy)newstate = RESTING;
	break; 
    case RESTING:
	if(totstatetime>=myres.minidle)
	    newstate = ignore?IDLING:WORKING;
	else if(!ignore){
	    laststatetime += totstatetime;
	    updatetimew(WORKING,laststatetime);
	    totstatetime = 0;	/* restart timer: still typing! */
	}
	break;
    }
    if(newstate!=currentstate){
	laststatetime = totstatetime;
	updatetimew(currentstate,laststatetime);
	setstate(newstate);
	totstatetime = 0;
    }
}
/*
 * change state on screen.
 */
setstate(newstate)
int newstate;
{
    int colour;
    Arg args[10];
    Cardinal n;
	
    switch(newstate){
    case WORKING:
	colour = myres.busyfg;
	break;
    case RESTING:
	colour = myres.recoverfg;
	break;
    case IDLING:
	colour = myres.idlefg;
	break;
    }
    if(myres.tomap)
	if(newstate==RESTING)XMapRaised(XtDisplay(toplevel),XtWindow(toplevel));
	else{
	    if(!myres.tounmap)
		XLowerWindow(XtDisplay(toplevel),XtWindow(toplevel));
	    else if(newstate==IDLING)
		XIconifyWindow(XtDisplay(toplevel),XtWindow(toplevel),DefaultScreen(XtDisplay(toplevel)));

	}
    currentstate = newstate;
    n = 0;
    XtSetArg(args[n], XtNbackground, colour); n++;
    XtSetValues(form, args, n);
    XtSetValues(label, args, n);
}
/*
 * update time shows in widget for last state
 */
updatetimew(oldstate,oldtime)
int oldstate,oldtime; 
{
    Arg args[10];
    Cardinal n;
    char str[100];

    oldtime /= 1000;
    sprintf(str,"%s %d:%02d",statename[oldstate],(int)(oldtime/60),(int)(oldtime%60));
    n = 0;
    XtSetArg(args[n], XtNlabel, str); n++;
    XtSetValues(timew, args, n);
}
/*
 * update total time typing
 */
updatetypingtime(time)
int time; 
{
    Arg args[10];
    Cardinal n;
    char str[100];
    static char *title;

    if(title==0){
	n = 0;
	XtSetArg(args[n], XtNlabel, &title); n++;
	XtGetValues(tottimew, args, n);
	if(title)title = XtNewString(title);
	else title = "";
    }
    if(time>=60*60)
    sprintf(str,"%s %d:%02d:%02d",title,(time/(60*60)),((time%(60*60))/60),(time%60));
    else sprintf(str,"%s %d:%02d",title,(time/60),(time%60));
    n = 0;
    XtSetArg(args[n], XtNlabel, str); n++;
    XtSetValues(tottimew, args, n);
}
/*
 * update total keystrokes and/or mouse
 */
updatekeystrokes(tot)
int tot; 
{
    Arg args[10];
    Cardinal n;
    char str[100];

    static char *title;

    if(title==0){
	n = 0;
	XtSetArg(args[n], XtNlabel, &title); n++;
	XtGetValues(keystrokesw, args, n);
	if(title)title = XtNewString(title);
	else title = "";
    }
    sprintf(str,"%s %d",title,tot);
    n = 0;
    XtSetArg(args[n], XtNlabel, str); n++;
    XtSetValues(keystrokesw, args, n);
}
/*
 * does not do root. assumed already done
 */
selectAllWindows(top)
Window top;
{
    XWindowAttributes attr;
    Window *child, dummy;
    unsigned int children, i, e, y;

    if (XQueryTree(display, top, &dummy, &dummy, &child, &children)==0)
        pexit("XQueryTree failed");
    for (i=0; i<children; i++)
	if(XGetWindowAttributes(display, child[i], &attr)){
	    e = attr.all_event_masks;
	    y = attr.your_event_mask;
	    e = (y^e)&WANTEDEVENTS;
	    if(e)
		XSelectInput(display, child[i], y^e);	/* even if 0! */
	    /* else already done */
	    selectAllWindows(child[i]);
	}
    if(child)XFree((char *)child);
}
#ifdef USEXTESTEXTN1
interpxtestaction(eventp,ignorep)
XEvent *eventp;
int *ignorep;
{
    XTestInputActionEvent *ep = (XTestInputActionEvent*)eventp;
    int act = ep->actions[0];
    int timepassed;

    /* ep->window ep->display */
    switch(act&XTestACTION_TYPE_MASK){
    case XTestKEY_ACTION:{
	XTestKeyInfo *kp;
	int code,kbd;
	kbd = XTestUnpackDeviceID(act);
	if(kbd)actstr = "Tkey";
	else actstr = "Tmouse";
	/* up = (act&XTestKEY_STATE_MASK)==XTestKEY_UP; */
	kp = (XTestKeyInfo*)ep->actions;
	code = kp->keycode;
	timepassed = kp->delay_time;
	break;
    }
    case XTestMOTION_ACTION:{
	XTestMotionInfo *mp;
	actstr = "Tmotion";
	mp = (XTestMotionInfo*)ep->actions;
	timepassed = mp->delay_time;
	*ignorep = TRUE;
	break;
    }
    case XTestJUMP_ACTION:{
	XTestJumpInfo *jp;
	actstr = "Tjump";
	jp = (XTestJumpInfo*)ep->actions;
	timepassed = jp->delay_time;
	*ignorep = TRUE;
	break;
    }
    case XTestDELAY_ACTION:{
	XTestDelayInfo *dp;
	actstr = "Tdelay";
	dp = (XTestDelayInfo*)ep->actions;
	timepassed = dp->delay_time;
	*ignorep = TRUE;
	break;
    }
    default:	/* shouldnt happen */
	actstr = "T?";
	timepassed = 0;
	*ignorep = TRUE;
    }
    return timepassed;
}
#endif
pexit(str)
char *str;
{
	if(str)perror(str);
	exit(str!=0);
}
