.\" restwatch man
.TH RESTWATCH 1X
.SH NAME
restwatch \- watches keystrokes and alerts for a rest after a time
.SH SYNOPSIS
.B
restwatch
'in \n(.iu+\w'\fBrestwatch \fR'u
.RB [ -busy
.IR busytime ]
.RB [ -idle
.IR idletime ]
.RB [ -verbose ]
.br
.RB [ -nomap ]
.RB [ -unmap ]
.RB [ -noext ]
.br
.RB [ -keystrokes
.IR count ]
.br
.RB [ -continuous
.IR msecs ]
.br
.RB [ -log
.IR minutes ]
.in
.SH DESCRIPTION
.B restwatch
is intended to run permanently during an X11 session and to monitor the
time during which the keyboard or mouse is used.
After a given time, a window pops up with a given position and size
and remains until no more typing is done for a given time.
Statistics of the total number of keystrokes and total typing time are
permanently displayed, and are written to stdout after a set amount of idleness
when they are also reset to zero.

The program can use one of two methods for monitoring keystrokes.

Using XTestExtension1.
This is an optional extension that has to exist on the client side when the
program was compiled, and in the server when the program is run.
It is the most efficient method to use, as all keystrokes, and mouse use, is
directly sent to the program.

If the above method is not supported, the window tree on the server is
periodically sampled and keypress events are selected on all
windows that already have some other clients selecting to receive keypress
events from them.
This method cannot notice use of the mouse buttons and is less efficient.
.SH OPTIONS
.B restwatch
uses the following resources:
.TP
verbose
Boolean, default false, print debug messages for each event.
.TP
map
Boolean, default true, pop the window up when you should be resting.
.TP
unmap
Boolean, default false, iconify the window when idle, else lower it to the
bottom of the window stack.
.TP
ext
Boolean, default true, use the XTestExtension1 if supported.
.TP
busy
Integer seconds, default 180, time to allow keyboard to be used.
This is not a recommended value.
.TP
idle
Integer seconds, default 30, time to wait while resting.
This is not a recommended value.
.TP
keystrokes
Integer count of keystrokes, default 100, to elapse before updating the
displayed total.
.TP
continuous
Integer milliseconds, default 2000, time between keystrokes that still
allows them to be considered as continuous typing.
This is used to display total typing time.
If you type slower than this you are not typing at all as far as this total
is concerned.
It is just for the statistics and does not affect the total busy time.
.TP
log
Integer minutes, default 6 hours, time of idleness after which the current
statistics will be printed and reset.
.TP
busyFg
Colour, default yellow, colour of window when busy.
.TP	
idleFg	
Colour, default black, colour of window when idle.
.TP	
recoverFg
Colour, default red, colour of window when resting.

The program has a class name of "Restwatch" and consists of the
widget hierarchy

   form Form
      time Label
      tottime Label
      keystrokes Label
      label Label

The above resources can be set as usual in an .Xdefaults file, for example, 
or on the command line by the usual X toolkit method:
.sp
	-xrm '*<resource>:<value>'

for example

	-xrm '*label:STOP!'
.sp
or with the options given in the synopsis.
.B restwatch
also accepts the usual X toolkit options (-display, etc).
This can be used to set the size and position of the notifying window, for
example, 
.sp
	-geometry 200x500-1+1
.SH AUTHOR
Mark M Martin. CETIA France. 95/06/06. mmm@cetia.fr
